#include <SPI.h>
#include <SdFat.h>
#include <TimerOneThree.h>

#define LCD_CS 0
#define CARD_CS 1
#define YM_CS 2
#define SN_CS 3
#define SN_RDY 16
#define SN_WE 17
#define YM_IRQ 18
#define YM_IC 19
#define YM_A0 20
#define YM_A1 21
#define YM_RD 22
#define YM_WR 23

//#define DEBUG
#ifdef DEBUG
#define DEBUG_PRINT(...)    Serial.print(__VA_ARGS__)
#define DEBUG_PRINTLN(...)  Serial.println(__VA_ARGS__)
#else
#define DEBUG_PRINT(...)
#define DEBUG_PRINTLN(...)
#endif

#define BUFFER_SIZE  1000
struct Buffer {
  uint8_t data[BUFFER_SIZE];
  uint8_t newest_index;
  uint8_t oldest_index;
};

enum BufferStatus {BUFFER_OK, BUFFER_EMPTY, BUFFER_FULL};
volatile struct Buffer sidBuffer = {{}, 0, 0};

enum BufferStatus bufferWrite(volatile struct Buffer *buffer, uint8_t byte) {
  uint8_t next_index = (((buffer->newest_index) + 1) % BUFFER_SIZE);
  if (next_index == buffer->oldest_index) {
    return BUFFER_FULL;
  }
  buffer->data[buffer->newest_index] = byte;
  buffer->newest_index = next_index;
  return BUFFER_OK;
}

enum BufferStatus bufferRead(volatile struct Buffer *buffer, uint8_t *byte) {
  if (buffer->newest_index == buffer->oldest_index) {
    return BUFFER_EMPTY;
  }
  *byte = buffer->data[buffer->oldest_index];
  buffer->oldest_index = ((buffer->oldest_index + 1) % BUFFER_SIZE);
  return BUFFER_OK;
}

SdFat SD;
StdioStream vgmStream;

byte pcmBuffer [ 7000 ];
uint32_t pcmBufferPosition;

void pushToSid() {//timer1 interrupt ~44.1kHz to send audio data (it is really 44.199kHz)
  byte instruction;
  while ( bufferRead(&sidBuffer, &instruction) == BUFFER_OK ) {
    byte addr;
    byte data;
    switch ( instruction ) {
      case 0x00:
        bufferRead( &sidBuffer, &data );
        PORTA = data;
        PORTB &= ~(0x08); //digitalWrite( SN_CS , LOW );
        PORTC &= ~(0x02); //digitalWrite( SN_WE , LOW );
        while ( !( PINC & 0x01 )); //while ( !digitalRead( SN_RDY ));
        PORTB |= 0x08; //digitalWrite( SN_CS , HIGH );
        PORTC |= 0x02; //digitalWrite( SN_WE , HIGH );
        break;
      case 0x01:
        PORTC &= ~(0x20);
        bufferRead( &sidBuffer, &addr );
        bufferRead( &sidBuffer, &data );
        PORTC &= ~(0x10); //digitalWrite( YM_A0 , LOW );
        PORTB &= ~(0x04); //digitalWrite( YM_CS , LOW );
        PORTA = addr;
        PORTC &= ~(0x80); //digitalWrite( YM_WR , LOW );
        delayMicroseconds(1);
        PORTC |= 0x80; //digitalWrite( YM_WR , HIGH );
        PORTB |= 0x04; //digitalWrite( YM_CS , HIGH );
        delayMicroseconds(1);
        PORTC |= 0x10; //digitalWrite( YM_A0 , HIGH );
        PORTB &= ~(0x04); //digitalWrite( YM_CS , LOW );
        PORTA = data;
        PORTC &= ~(0x80); //digitalWrite( YM_WR , LOW );
        delayMicroseconds(1);
        PORTC |= 0x80; //digitalWrite( YM_WR , HIGH );
        PORTB |= 0x04; //digitalWrite( YM_CS , HIGH );
        break;
      case 0x02:
        PORTC |= 0x20;
        bufferRead( &sidBuffer, &addr );
        bufferRead( &sidBuffer, &data );
        PORTC &= ~(0x10); //digitalWrite( YM_A0 , LOW );
        PORTB &= ~(0x04); //digitalWrite( YM_CS , LOW );
        PORTA = addr;
        PORTC &= ~(0x80); //digitalWrite( YM_WR , LOW );
        delayMicroseconds(1);
        PORTC |= 0x80; //digitalWrite( YM_WR , HIGH );
        PORTB |= 0x04; //digitalWrite( YM_CS , HIGH );
        delayMicroseconds(1);
        PORTC |= 0x10; //digitalWrite( YM_A0 , HIGH );
        PORTB &= ~(0x04); //digitalWrite( YM_CS , LOW );
        PORTA = data;
        PORTC &= ~(0x80); //digitalWrite( YM_WR , LOW );
        delayMicroseconds(1);
        PORTC |= 0x80; //digitalWrite( YM_WR , HIGH );
        PORTB |= 0x04; //digitalWrite( YM_CS , HIGH );
        break;
      case 0x03:
        break;
    }
  }
}

void delaySample() {
  bufferWrite( &sidBuffer, 0x03 );
}

void snSend( byte data ) {
  bufferWrite( &sidBuffer, 0x00 );
  bufferWrite( &sidBuffer, data );
}

void ymSend( byte port, byte addr, byte data ) {
  bufferWrite( &sidBuffer, port + 1 );
  bufferWrite( &sidBuffer, addr );
  bufferWrite( &sidBuffer, data );
}

void seekPcmPointer() {
  pcmBufferPosition = 0;
  for ( int i = 0; i < 4; i++ ) pcmBufferPosition += uint32_t( vgmStream.getc()) << ( 8 * ( i ) );
  DEBUG_PRINT( "Seeking PCM Pointer to Offset " );
  DEBUG_PRINTLN( pcmBufferPosition, HEX );
}

void playSample () {
  DEBUG_PRINT( "YM Play Sample at PCM Address 0x" );
  DEBUG_PRINT( pcmBufferPosition, HEX );
  DEBUG_PRINT( " Data: 0x" );
  DEBUG_PRINTLN( pcmBuffer [ pcmBufferPosition ], HEX );
  ymSend( 0, 0x2A, pcmBuffer [ pcmBufferPosition++ ] );
}

long parseCommand( byte command ) { //return number of bytes skipped
  byte addr;
  byte data;
  byte temp;
  uint32_t s = 0;
  switch ( command ) {
    case 0x4F: // should I just write this to the SN?
      DEBUG_PRINT( "SN Data 0x" );
      data = vgmStream.getc(); //get next byte from stream
      DEBUG_PRINTLN( data, HEX );
      snSend( data );
      return 1;
    case 0x50:
      DEBUG_PRINT( "SN Data 0x" );
      data = vgmStream.getc(); //get next byte from stream
      DEBUG_PRINTLN( data, HEX );
      snSend( data );
      return 1;
    case 0x52:
      DEBUG_PRINT( "YM 0 Address 0x" );
      addr = vgmStream.getc();
      DEBUG_PRINT( addr, HEX );
      DEBUG_PRINT( " Data 0x" );
      data = vgmStream.getc();
      DEBUG_PRINTLN( data, HEX );
      ymSend( 0, addr, data );
      return 2;
    case 0x53:
      DEBUG_PRINT( "YM 1 Address 0x" );
      addr = vgmStream.getc();
      DEBUG_PRINT( addr, HEX );
      DEBUG_PRINT( " Data 0x" );
      data = vgmStream.getc();
      DEBUG_PRINTLN( data, HEX );
      ymSend( 1, addr, data );
      return 2;
    case 0x67:
      DEBUG_PRINT( "Data Block Type " );
      temp = vgmStream.getc(); //0x66
      temp = vgmStream.getc(); //Data Block Type
      pcmBufferPosition = vgmStream.ftell();
      DEBUG_PRINT( String( temp ));
      for ( int i = 0; i < 4; i++ ) s += ( uint32_t( vgmStream.getc() ) << ( 8 * i ));
      DEBUG_PRINT( " Space needed:" );
      for ( uint32_t i = 0; i < s; i++ ) pcmBuffer[ i ] = vgmStream.getc();
      DEBUG_PRINT( " Seeking to " );
      //vgmStream.fseek( 0x03D6, SEEK_CUR );
      DEBUG_PRINTLN( vgmStream.ftell(), HEX );
      return s;
    case 0x80:
    case 0x81:
    case 0x82:
    case 0x83:
    case 0x84:
    case 0x85:
    case 0x86:
    case 0x87:
    case 0x88:
    case 0x89:
    case 0x8A:
    case 0x8B:
    case 0x8C:
    case 0x8D:
    case 0x8E:
    case 0x8F:
      playSample();
      s = command & 0x0F;
      while ( s-- > 0 ) delaySample();
      return 1;
    case 0x61:
      DEBUG_PRINT( "Delaying For " );
      for ( int i = 0; i < 2; i++ ) s += ( uint32_t( vgmStream.getc() ) << ( 8 * i ));
      DEBUG_PRINTLN( String( s ) + " samples" );
      while ( s-- > 0 ) delaySample();
      return 2;
    case 0x70:
    case 0x71:
    case 0x72:
    case 0x73:
    case 0x74:
    case 0x75:
    case 0x76:
    case 0x77:
    case 0x78:
    case 0x79:
    case 0x7A:
    case 0x7B:
    case 0x7C:
    case 0x7D:
    case 0x7E:
    case 0x7F:
      DEBUG_PRINT( "Delaying " );
      DEBUG_PRINTLN( command & 0x0F );
      s = command & 0x0F;
      while ( s-- > 0 ) delaySample();
      return 0;
    case 0xE0:
      seekPcmPointer();
      return 4;
    case 0x66:
      DEBUG_PRINT( "End of Sound Data" );
      vgmStream.fclose();
      return 0;
    default:
      DEBUG_PRINT( "Unknown Command: 0x" );
      DEBUG_PRINTLN( command, HEX );
      return 0;
  }
  return 0;
}

void setup() {
  SD.begin( CARD_CS );
  Serial.begin( 115200 );
  //SN76289 Init
  DDRA = 0xFF;
  PORTA = 0x00;
  pinMode( SN_RDY, INPUT_PULLUP );
  pinMode( SN_CS, OUTPUT );
  pinMode( SN_WE, OUTPUT );
  digitalWrite( SN_CS , LOW );
  digitalWrite( SN_WE , LOW );
  //YM 2612 Init
  pinMode( YM_CS, OUTPUT );
  pinMode( YM_IC, OUTPUT );
  pinMode( YM_A0, OUTPUT );
  pinMode( YM_A1, OUTPUT );
  pinMode( YM_RD, OUTPUT );
  pinMode( YM_WR, OUTPUT );
  pinMode( YM_IRQ, INPUT_PULLUP );
  digitalWrite( YM_A0 , LOW );
  digitalWrite( YM_A1 , LOW );
  digitalWrite( YM_CS , HIGH );
  digitalWrite( YM_WR , HIGH );
  digitalWrite( YM_RD , HIGH );
  digitalWrite( YM_IC , HIGH );
  delay(10);
  digitalWrite( YM_IC , LOW );
  delay(10);
  digitalWrite( YM_IC , HIGH );
  delay(10);
}
int freeRam ()
{
  extern int __heap_start, *__brkval;
  int v;
  return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval);
}
void loop() {
  DEBUG_PRINTLN( "Opening VGM File" );
  if ( vgmStream.fopen( "SONIC.VGM", "r" ) ) {
    DEBUG_PRINTLN( "VGM File Open" );
    uint32_t ident = 0;
    for ( int i = 0; i < 4; i++ ) ident += uint32_t(vgmStream.getc()) << ( 8 * i );
    if ( ident != 0x206d6756 ) {
      DEBUG_PRINT( "Ident Match Failed!" );
      return;
    }
    uint32_t eof = 0;
    for ( int i = 0; i < 4; i++ ) eof += uint32_t(vgmStream.getc()) << ( 8 * i );
    DEBUG_PRINTLN( "EOF:" + String( eof ) );
    DEBUG_PRINTLN( "Header Valid" );
    vgmStream.fseek( 0x40, SEEK_SET ); // cheat for now, need to read real VGM offset in header
    DEBUG_PRINTLN( "--------------VGM File Data Start--------------" );
    while ( !vgmStream.feof() ) {
      byte command = vgmStream.getc();
      DEBUG_PRINT( "Ram:" + String( freeRam()) );
      DEBUG_PRINT( " Addr:0x" );
      DEBUG_PRINT( vgmStream.ftell(), HEX );
      DEBUG_PRINT( " Cmd:0x" );
      DEBUG_PRINT( command, HEX );
      DEBUG_PRINT( " --> " );
      if ( command == 0x66 ) break;
      parseCommand( command );
      pushToSid();
    }
    DEBUG_PRINTLN( "VGM File Ended" );
    vgmStream.fclose();
  }
}
